/**************************************************************************/
/*  lnxsound.c  Version 1.0                                               */
/*  Copyright (C) 2007 Mariano Alvarez Fernandez                          */
/*  Lea en el fichero leeme.txt los detalles                              */
/*  See readme.txt for details                                            */
/**************************************************************************/
/*  Rutinas de sonido en linux usando /dev/dsp para consola y X11         */
/**************************************************************************/

#include <linux/soundcard.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <math.h>
#include <mgrx.h>
#include "grsound.h"

static int sound_inited = 0;
static int sound_enabled = 0;
static int handle;
static int octave = 3;
static int estacato = 1;
static int volumen = 32;

static unsigned frecuencia[] = {
         33,  35,  37,  39,  41,  44,  46,  49,  52,  55,  58,  62,
         65,  69,  73,  78,  82,  87,  92,  98, 104, 110, 117, 123,
        131, 139, 147, 156, 165, 175, 185, 196, 208, 220, 233, 247,
        262, 277, 294, 311, 330, 349, 370, 392, 415, 440, 466, 494,
        523, 554, 587, 622, 659, 698, 740, 784, 831, 880, 932, 988,
       1046,1109,1175,1245,1329,1397,1480,1568,1661,1760,1865,1976,
       2093,2217,2349,2489,2637,2794,2960,3136,3322,3520,3729,3951 };

#define BUFSIZE 8000
static unsigned char buffer[BUFSIZE];

static int generatetone(int freq, int msec)
{
  float a, x, tic;
  int b, nsamples;

  if (msec > BUFSIZE/8) msec = BUFSIZE/8;
  tic = (freq * 3.141592654) / 4000;
  nsamples = msec * 8;
  x = 0.0;

  for (b=0; b<nsamples;b++) {
    /* onda cuadrada */
    a=sin(x);
    if (a > 0) buffer[b] = 127 + volumen;
    else buffer[b] = 127 - volumen;
//    /* onda senoidal */
//    a=sin(x);
//    a *= volumen;
//    buffer[b]=(unsigned char)a;
//    buffer[b]^=0x80;
    x += tic;
  };

  return nsamples;
};

int GrSoundIni(void)
{
  int a, ssf, error;

  if ((handle=open("/dev/dsp", O_WRONLY)) < 0)
    return 0;

  error = 0;
  a = AFMT_U8;  // 8 bit unsigned samples
  if (ioctl(handle, SNDCTL_DSP_SETFMT, &a) < 0) error -=1;
  if (a != AFMT_U8) error -=1;
  a = 8000;  // 8 kHz
  if (ioctl(handle, SNDCTL_DSP_SPEED, &a) < 0) error -=1;
  ssf = (4 << 16) | 5;  // buffer: 4 fragments of 32 bytes
  if (ioctl(handle,SNDCTL_DSP_SETFRAGMENT,&ssf) < 0) error -=1;

  if (error < 0) {
    close(handle);
    return 0;
  }

  sound_inited = 1;
  sound_enabled = 1;

  return 1;
}

void GrSoundEnd(void)
{
  if (sound_inited) {
    close(handle);
  }

  sound_inited = 0;
  sound_enabled = 0;
}

void GrSoundEnable(void)
{
  if (sound_inited) {
    sound_enabled = 1;
  }
}

void GrSoundDisable(void)
{
  sound_enabled = 0;
}

int GrSoundIsEnabled(void)
{
  return sound_enabled;
}

int GrSoundSetVolume(int vol)
{
  int vol_ant;

  vol_ant = volumen;
  if (vol < 0) vol = 0;
  if (vol > 127) vol = 127;
  volumen = vol;
  return vol_ant;
}

int GrSoundSetOctave(int oct)
{
  int oct_ant;

  oct_ant = octave;
  if( oct > 0 && oct < 7 )
    octave = oct;
  return oct_ant;
}

int GrSoundSetEstacato(int est)
{
  int est_ant;

  est_ant = estacato;
  estacato = est;
  return est_ant;
}

void GrSoundPlayMelodia(struct NOTA *mel, int nnotas, int checkstop)
{
  int i;

  for( i=0; i<nnotas; i++){
    GrSoundPlayNote(mel->nota, mel->t);
    mel++;
    if( checkstop && GrEventCheck() ){
      return;
      }
    }
}

void GrSoundPlayNote(int note, int t)
{
  if( note != SILENCIO ){
    GrSoundPlaySound((octave - 1)*12+note, t);
    if( estacato ){
      GrSleep( 10 );
      }
    }
  else{
    GrSleep( t );
    }
}

void GrSoundPlaySound(int numsound, int t)
{
  int nsamples, a = 0;

  if (sound_enabled) {
    if (numsound < 0) numsound = 0;
    if (numsound > 83) numsound = 83;
    nsamples = generatetone(frecuencia[numsound], t);
    write(handle, buffer, nsamples);
    GrSleep(1);
    ioctl(handle, SNDCTL_DSP_SYNC, &a);
  }
  else {
    GrSleep(t);
  }
}
