/**************************************************************************/
/*  parchit.c Version 4.1 Modulo de PARCHIS.C                             */
/*  Copyright (C) 1992, 1999, 2007, 2017 Mariano Alvarez Fernandez        */
/*  Lea en el fichero LEEME.TXT los detalles                              */
/*  See README.TXT for details                                            */
/**************************************************************************/
/*  Rutinas del juego en si                                               */
/**************************************************************************/

#include <stdlib.h>
#include "parchit.h"

/**************************************************************************/

void inicia_tablero( struct Tablero *tbl )
{
  int i, j;

  for( i=0; i<68; i++ ){
    tbl->corredor[i].tipo = NORMAL;
    tbl->corredor[i].f[0] = VACIO;
    tbl->corredor[i].f[1] = VACIO;
    }
  tbl->corredor[4].tipo   = ENTRADA_AMAR;
  tbl->corredor[11].tipo  = SEGURO;
  tbl->corredor[16].tipo  = SALIDA;
  tbl->corredor[21].tipo  = ENTRADA_AZUL;
  tbl->corredor[28].tipo  = SEGURO;
  tbl->corredor[33].tipo  = SALIDA;
  tbl->corredor[38].tipo  = ENTRADA_ROJO;
  tbl->corredor[45].tipo  = SEGURO;
  tbl->corredor[50].tipo  = SALIDA;
  tbl->corredor[55].tipo  = ENTRADA_VERD;
  tbl->corredor[62].tipo  = SEGURO;
  tbl->corredor[67].tipo  = SALIDA;

  for( i=0; i<4; i++ )
    for( j=0; j<7; j++ ){
      tbl->llegada[i][j].f[0] = VACIO;
      tbl->llegada[i][j].f[1] = VACIO;
    }

  for( i=AMAR; i<=VERD; i++ )
    for( j=0; j<4; j++ )
      tbl->meta[i].f[j] = VACIO;

  for( i=AMAR; i<=VERD; i++ )
    for( j=0; j<4; j++ ){
      tbl->casa[i].f[j] = i;
      tbl->casa[i].nf[j] = j;
      }

  for( i=AMAR; i<=VERD; i++ )
    for( j=0; j<4; j++ ){
      tbl->jugador[i].ha_salido = 0;
      tbl->jugador[i].ficha[j].lugar = CASA;
      tbl->jugador[i].ficha[j].pos = 0;
      tbl->jugador[i].ficha[j].ind = j;
      }

  for( i=AMAR; i<=VERD; i++ ){
    tbl->resumen[i].nfcasa = 4;
    tbl->resumen[i].nftbl  = 0;
    tbl->resumen[i].nfmeta = 0;
    }

  tbl->turno = AMAR;
  tbl->siguiente_final = 0;
  for( i=0; i<4; i++ )
    tbl->orden_final[i] = -1;
}

/**************************************************************************/

void inicia_jugadores( struct Tablero *tbl, int jamar, int jazul,
                       int jrojo, int jverd )
{
  int i;

  tbl->jugador[AMAR].tipo = jamar;
  tbl->jugador[AZUL].tipo = jazul;
  tbl->jugador[ROJO].tipo = jrojo;
  tbl->jugador[VERD].tipo = jverd;
  tbl->num_jugadores = 0;
  for( i=AMAR; i<=VERD; i++ )
    if( tbl->jugador[i].tipo != NOJUEGA ) tbl->num_jugadores++;
}

/**************************************************************************/

int dado( void )
{
  return (random() % 6) + 1;
}

/**************************************************************************/

int hace6a7( struct Tablero *tbl, int cjugador )
{
  if( tbl->resumen[cjugador].nfcasa == 0 ) return 1;
  return 0;
}

/**************************************************************************/

int dadocorr( struct Tablero *tbl, int cjugador )
{
  int i;

  i = dado();
  if( i == 6 && hace6a7( tbl,cjugador ) ) i++;
  return i;
}

/**************************************************************************/

int tipo_jugador( struct Tablero *tbl, int cjugador )
{
  if( tbl->resumen[cjugador].nfmeta == 4 ) return NOJUEGA; /* ya ha llegado */
  return tbl->jugador[cjugador].tipo;
}

/**************************************************************************/

int genera_jugada( struct Tablero *tbl, int cjugador, int nficha,
                   int vdado, struct Jugada *jugada )
{
  static int ent[4] = { 4,21,38,55 };
  static int sal[4] = { 67,16,33,50 };
  struct Posicion *posini, *posfin;
  struct Casillac *casillac;
  int i, tipo;

  jugada->cjugador = cjugador;
  jugada->nficha = nficha;
  jugada->vdado = vdado;
  jugada->cjugdcom = VACIO;
  jugada->nfichcom = 0;
  posini = &(tbl->jugador[cjugador].ficha[nficha]);
  posfin = &(jugada->destino);
  jugada->origen = *posini;
  *posfin = *posini;
  inicia_analisis( &(jugada->anal) );

  if( posini->lugar == META ) return 0;

  /* siempre se supone que las casillas se rellenan por orden, no
     puede haber una ficha en ind=1 si no la hay en ind=0 */
  /* solo sale de casa si se saco un 5 y no hay dos fichas propias en
     la salida, si hay 2 fichas y una no es propia, ocupa su puesto */

  if( posini->lugar == CASA ){
    if( vdado != 5 ) return 0;  /* no vale */
    jugada->anal.entra = 1;
    casillac = &( tbl->corredor[ent[cjugador]] );
    if( casillac->f[0] == cjugador &&
	casillac->f[1] == cjugador ) return 0;
    posfin->lugar = CORREDOR;
    posfin->pos = ent[cjugador];
    posfin->ind = 0;
    if( casillac->f[0] == VACIO )
      return 1;
    if( casillac->f[1] == VACIO ){
      posfin->ind = 1;
      return 1;
      }
    jugada->anal.come = 1;
    if( casillac->f[1] != cjugador ){
      jugada->cjugdcom = tbl->corredor[posfin->pos].f[1];
      jugada->nfichcom = tbl->corredor[posfin->pos].nf[1];
      posfin->ind = 1;
      return 1;
      }
    jugada->cjugdcom = tbl->corredor[posfin->pos].f[0];
    jugada->nfichcom = tbl->corredor[posfin->pos].nf[0];
    return 1;
    }

  if( posini->lugar == CORREDOR ){
    /* Mira si rompe puente */
    if( tbl->corredor[posini->pos].f[0] == tbl->corredor[posini->pos].f[1] )
      jugada->anal.rpte = 1;
    /* Mira si deja seguro */
    tipo = tbl->corredor[posini->pos].tipo;
    if( tipo == SEGURO || tipo == SALIDA )
      jugada->anal.dseg = 1;
    if( tipo >= ENTRADA ){
      i = tipo - ENTRADA;
      if( i == cjugador ||
          tbl->jugador[i].tipo == NOJUEGA ||
          tbl->resumen[i].nfcasa == 0 ||
          (tbl->resumen[i].nfcasa < 4 &&
             tbl->corredor[posini->pos].f[1] == VACIO) )
	jugada->anal.dseg = 1;
      }
    /* Mueve */
    for( i=0; i<vdado; i++ ){
      if( posfin->pos == sal[cjugador] ){
	jugada->anal.sale = 1;
	posfin->lugar = LLEGADA;
	posfin->pos = -1; /* posici�n ficticia */
	vdado -= i;
	goto EN_LLEGADA;
	}
      posfin->pos++;
      if( posfin->pos > 67 ) posfin->pos = 0;
      if( tbl->corredor[posfin->pos].f[1] != VACIO ){
	/* hay una barrera, pero puede ser de fichas distintas en seguro */
	/* de todas formas, si es la posicion final, no vale */
	if( i == vdado - 1 ) return 0; /* no vale */
	if( tbl->corredor[posfin->pos].f[0] ==
	    tbl->corredor[posfin->pos].f[1] ) return 0; /* no vale */
	}
      }
    /* estamos en la posicion final, ver si estamos solos o si se come */
    /* o si formamos barrera, o si estamos en seguro y no se come */
    tipo = tbl->corredor[posfin->pos].tipo;
    if( tipo == SEGURO || tipo == SALIDA )
      jugada->anal.aseg = 1;
    if( tipo >= ENTRADA ){
      i = tipo - ENTRADA;
      if( i == cjugador ||
          tbl->jugador[i].tipo == NOJUEGA ||
          tbl->resumen[i].nfcasa == 0 ||
          (tbl->resumen[i].nfcasa < 4 &&
             tbl->corredor[posfin->pos].f[0] == VACIO) )
	jugada->anal.aseg = 1;
      }
    if( tbl->corredor[posfin->pos].f[0] == VACIO ){
      posfin->ind = 0;
      return 1; /* ok, s�la */
      }
    if( tbl->corredor[posfin->pos].f[0] == cjugador ){
      posfin->ind = 1;
      jugada->anal.hpte = 1;
      return 1; /* ok,  hace puente */
      }
    if( tbl->corredor[posfin->pos].tipo != NORMAL ){
      posfin->ind = 1;
      return 1; /* ok, en seguro con otra */
      }
    posfin->ind = 0;
    jugada->anal.come = 1;
    jugada->cjugdcom = tbl->corredor[posfin->pos].f[0];
    jugada->nfichcom = tbl->corredor[posfin->pos].nf[0];
    return 1; /* ok, come */
    }

EN_LLEGADA:
  if( posfin->pos + vdado > 7 ) return 0; /* no vale */
  if( posfin->pos + vdado == 7 ){
    posfin->lugar = META;
    jugada->anal.meta = 1;
    for( i=0; tbl->meta[cjugador].f[i] != VACIO; i++ );
    posfin->ind = i;
    return 1; /* ok, en la meta */
    }
  posfin->pos += vdado;
  if( tbl->llegada[cjugador][posfin->pos].f[1] != VACIO ) return 0; /* no vale */
  posfin->ind = 0;
  if( tbl->llegada[cjugador][posfin->pos].f[0] != VACIO ) posfin->ind = 1;
  return 1; /* ok */
}

/**************************************************************************/

int genera_jugadas_posibles( struct Tablero *tbl,
                   struct JugadasPosibles *jpo, int vdado, int cjugador )
{
  int i, sale, rpte, njc, nj = 0;

  for( i=0; i<4; i++ ){
    nj += genera_jugada( tbl,cjugador,i,vdado,&(jpo->jugada[nj]) );
    }
  /* mira si es 5 y sale de casa, en cuyo caso solo hay una jugada */
  /* permitida */
  if( vdado == 5 ){
    sale = -1;
    for( i=0; sale < 0 && i < nj; i++ )
      if( jpo->jugada[i].anal.entra ) sale = i;
    if( sale >= 0 ){
      if( i > 0 ) jpo->jugada[0] = jpo->jugada[sale];
      nj = 1;
      }
    }
  /* mira si es 6 o 7 y rompe puente, entonces es obligado */
  if( vdado == 6 || vdado == 7 ){
    rpte = 0;
    for( i=0; rpte == 0 && i < nj; i++ )
      if( jpo->jugada[i].anal.rpte ) rpte = 1;
    if( rpte ){
      njc = 0;
      for( i=0; i<nj; i++ )
        if( jpo->jugada[i].anal.rpte )
          jpo->jugada[njc++] = jpo->jugada[i];
      nj = njc;
      }
    }
  jpo->njugadas = nj;
  return nj;
}

/**************************************************************************/

int jugada_valida( struct JugadasPosibles *jpo, struct Jugada *jugada )
{
  int i;

  for( i=0; i<jpo->njugadas; i++ ){
    if( compara_jugada( jugada,&(jpo->jugada[i]) ) ) return 1;
    }
  return 0;
}

/**************************************************************************/

int compara_jugada( struct Jugada *j1, struct Jugada *j2 )
{
  if( j1->cjugador != j2->cjugador ) return 0;
  if( j1->nficha != j2->nficha ) return 0;
  return compara_posicion( &(j1->destino),&(j2->destino ) );
}

/**************************************************************************/

int compara_posicion( struct Posicion *p1, struct Posicion *p2 )
{
  if( p1->lugar != p2->lugar ) return 0;
  if( p1->pos != p2->pos ) return 0;
  if( p1->ind != p2->ind ) return 0;
  return 1;
}

/**************************************************************************/

int realiza_jugada( struct Tablero *tbl, struct Jugada *jugada )
{
  struct Posicion *p, paux;
  int ficha_comida, en_meta, jfin;

  /* actualiza resumen */
  if( jugada->anal.entra ){
    (tbl->resumen[jugada->cjugador].nfcasa)--;
    (tbl->resumen[jugada->cjugador].nftbl)++;
    }
  if( jugada->anal.meta ){
    (tbl->resumen[jugada->cjugador].nfmeta)++;
    (tbl->resumen[jugada->cjugador].nftbl)--;
    }
  /* libera la posici�n inicial */
  p = &(tbl->jugador[jugada->cjugador].ficha[jugada->nficha]);
  if( p->lugar == CASA )
    vacia_casilla_casa( tbl,jugada->cjugador,p->ind );
  else if( p->lugar == CORREDOR )
    vacia_casilla_corredor( tbl,p->pos,p->ind );
  else /* no puede ser la META, luego es la LLEGADA */
    vacia_casilla_llegada( tbl,jugada->cjugador,p->pos,p->ind );

  /* coloca la nueva posicion en los datos de la ficha */
  *p = jugada->destino;

  /* coloca la ficha en la nueva posicion y mira si se come ficha */
  ficha_comida = 0;
  en_meta = 0;
  jfin = 0;
  if( p->lugar == CORREDOR ){
    ficha_comida = a_casa( tbl,p->pos,p->ind );
    tbl->corredor[p->pos].f[p->ind] = jugada->cjugador;
    tbl->corredor[p->pos].nf[p->ind] = jugada->nficha;
    }
  else if( p->lugar == LLEGADA ){
    tbl->llegada[jugada->cjugador][p->pos].f[p->ind] = jugada->cjugador;
    tbl->llegada[jugada->cjugador][p->pos].nf[p->ind] = jugada->nficha;
    }
  else{ /* debe ser la meta */
    tbl->meta[jugada->cjugador].f[p->ind] = jugada->cjugador;
    tbl->meta[jugada->cjugador].nf[p->ind] = jugada->nficha;
    if( tbl->resumen[jugada->cjugador].nfmeta < 4 )
      en_meta = 1;
    else
      jfin = 1;
    }

  /* si es el primer 5 hay que sacar dos fichas */
  if( jugada->anal.entra && !tbl->jugador[jugada->cjugador].ha_salido ){
    (tbl->resumen[jugada->cjugador].nfcasa)--;
    (tbl->resumen[jugada->cjugador].nftbl)++;
    paux = *p;
    paux.ind = (p->ind == 0) ? 1 : 0;
    vacia_casilla_casa( tbl,jugada->cjugador,0 );
    tbl->jugador[jugada->cjugador].ficha[1] = paux;
    ficha_comida = a_casa( tbl,paux.pos,paux.ind );
    tbl->corredor[paux.pos].f[paux.ind] = jugada->cjugador;
    tbl->corredor[paux.pos].nf[paux.ind] = 1;
    tbl->jugador[jugada->cjugador].ha_salido = 1;
    }

  /* devuelve valor de dado de la jugada */
  if( ficha_comida ) return 20;
  if( en_meta ) return 10;
  if( jfin ) return -1; /* indica que ya han llegado todas sus fichas */
  return 0;
}

/**************************************************************************/

void vacia_casilla_casa( struct Tablero *tbl, int cjugador, int ncasilla )
{
  int i;

  for( i=ncasilla; i<3; i++ ){
    tbl->casa[cjugador].f[i] = tbl->casa[cjugador].f[i+1];
    tbl->casa[cjugador].nf[i] = tbl->casa[cjugador].nf[i+1];
    if( tbl->casa[cjugador].f[i] != VACIO )
      (tbl->jugador[cjugador].ficha[tbl->casa[cjugador].nf[i]].ind)--;
    }
  tbl->casa[cjugador].f[3] = VACIO;
}

/**************************************************************************/

void vacia_casilla_corredor( struct Tablero *tbl, int pos, int ncasilla )
{
  int i, cj;

  for( i=ncasilla; i<1; i++ ){
    tbl->corredor[pos].f[i] = tbl->corredor[pos].f[i+1];
    tbl->corredor[pos].nf[i] = tbl->corredor[pos].nf[i+1];
    cj = tbl->corredor[pos].f[i];
    if( cj != VACIO )
      (tbl->jugador[cj].ficha[tbl->corredor[pos].nf[i]].ind)--;
    }
  tbl->corredor[pos].f[1] = VACIO;
}

/**************************************************************************/

void vacia_casilla_llegada( struct Tablero *tbl, int cjugador,
                            int pos, int ncasilla )
{
  int i;

  for( i=ncasilla; i<1; i++ ){
    tbl->llegada[cjugador][pos].f[i] = tbl->llegada[cjugador][pos].f[i+1];
    tbl->llegada[cjugador][pos].nf[i] = tbl->llegada[cjugador][pos].nf[i+1];
    if( tbl->llegada[cjugador][pos].f[i] != VACIO )
      (tbl->jugador[cjugador].ficha[tbl->llegada[cjugador][pos].nf[i]].ind)--;
    }
  tbl->llegada[cjugador][pos].f[1] = VACIO;
}

/**************************************************************************/

int a_casa( struct Tablero *tbl, int pos, int ncasilla )
{
  int i, cf, nf;

  cf = tbl->corredor[pos].f[ncasilla];
  nf = tbl->corredor[pos].nf[ncasilla];
  if( cf == VACIO ) return 0;

  /* actualiza resumen */
  (tbl->resumen[cf].nfcasa)++;
  (tbl->resumen[cf].nftbl)--;

  for( i=0; tbl->casa[cf].f[i] != VACIO; i++ );

  tbl->casa[cf].f[i] = cf;
  tbl->casa[cf].nf[i] = nf;
  tbl->jugador[cf].ficha[nf].lugar = CASA;
  tbl->jugador[cf].ficha[nf].pos = 0;
  tbl->jugador[cf].ficha[nf].ind = i;

  return 1;
}

/**************************************************************************/

void inicia_analisis( struct Analisis *anal )
{
  anal->entra = 0;
  anal->sale = 0;
  anal->meta = 0;
  anal->come = 0;
  anal->hpte = 0;
  anal->rpte = 0;
  anal->dseg = 0;
  anal->aseg = 0;
  anal->nficom = 0;
  anal->nfipel = 0;
  anal->nffcom = 0;
  anal->nffpel = 0;
  anal->ncfin = 0;
}

/**************************************************************************/

void tres_seises( struct Tablero *tbl, int cj, int nf, struct Jugada *jugada )
{
  int pos, ind;

  jugada->cjugador = VACIO;
  pos = tbl->jugador[cj].ficha[nf].pos;
  ind = tbl->jugador[cj].ficha[nf].ind;

  a_casa( tbl,pos,ind );
  vacia_casilla_corredor( tbl,pos,ind );
}

/**************************************************************************/

int paso_adelante( int cjugador, struct Posicion *posini,
                   struct Posicion *posfin )
{
  static int ent[4] = { 4,21,38,55 };
  static int sal[4] = { 67,16,33,50 };

  *posfin = *posini;
  posfin->ind = 0;

  if( posini->lugar == META ) return 0;
  if( posini->lugar == CASA ){
    posfin->lugar = CORREDOR;
    posfin->pos = ent[cjugador];
    return 1;
    }
  if( posini->lugar == CORREDOR ){
    if( posini->pos == sal[cjugador] ){
      posfin->lugar = LLEGADA;
      posfin->pos = 0;
      return 1;
      }
    posfin->pos++;
    if( posfin->pos > 67 ) posfin->pos = 0;
    return 1;
    }
  /* en LLEGADA */
  if( posini->pos == 6 ){
    posfin->lugar = META;
    posfin->pos = 0;
    return 1;
    }
  posfin->pos++;
  return 1;
}

/**************************************************************************/

int paso_atras( int cjugador, struct Posicion *posini,
                struct Posicion *posfin )
{
  static int ent[4] = { 4,21,38,55 };
  static int sal[4] = { 67,16,33,50 };

  *posfin = *posini;
  posfin->ind = 0;

  if( posini->lugar == CASA ) return 0;
  if( posini->lugar == META ){
    posfin->lugar = LLEGADA;
    posfin->pos = 6;
    return 1;
    }
  if( posini->lugar == CORREDOR ){
    if( posini->pos == ent[cjugador] ){
      posfin->lugar = CASA;
      posfin->pos = 0;
      return 1;
      }
    posfin->pos--;
    if( posfin->pos < 0 ) posfin->pos = 67;
    return 1;
    }
  /* en LLEGADA */
  if( posini->pos == 0 ){
    posfin->lugar = CORREDOR;
    posfin->pos = sal[cjugador];
    return 1;
    }
  posfin->pos--;
  return 1;
}

/**************************************************************************/

void analisis_global( struct Tablero *tbl, struct AnalGlobal *anal )
{
  int i, j, dado, nj;
  struct JugadasPosibles jpo;

  for( i=0; i<4; i++ )
    for( j=0; j<4; j++ ){
      anal->nficom[i][j] = 0;
      anal->nfipel[i][j] = 0;
      }
  for( i=0; i<4; i++ )
    if( tbl->jugador[i].tipo != NOJUEGA )
      for( dado=1; dado<7; dado++ ){
        if( dado == 6 && hace6a7( tbl,i ) ) dado++;
        nj = genera_jugadas_posibles( tbl,&jpo,dado,i );
        for( j=0; j<nj; j++ )
          if( jpo.jugada[j].anal.come && jpo.jugada[j].cjugdcom != VACIO ){
            anal->nficom[i][jpo.jugada[j].nficha]++;
            anal->nfipel[jpo.jugada[j].cjugdcom][jpo.jugada[j].nfichcom]++;
            }
        }
}

/**************************************************************************/

void completa_info_anl( struct Tablero *tbl, struct JugadasPosibles *jpo )
{
  static int ent[4] = { 4,21,38,55 };
  struct Tablero tblaux;
  struct AnalGlobal analg, analg2;
  int i, nj, nf, n;

  analisis_global( tbl,&analg );
  for( i=0; i<jpo->njugadas; i++ ){
    nj = jpo->jugada[i].cjugador;
    nf = jpo->jugada[i].nficha;
    jpo->jugada[i].anal.nficom = analg.nficom[nj][nf];
    jpo->jugada[i].anal.nfipel = analg.nfipel[nj][nf];
    tblaux = *tbl;
    realiza_jugada( &tblaux,&(jpo->jugada[i]) );
    analisis_global( &tblaux,&analg2 );
    jpo->jugada[i].anal.nffcom = analg2.nficom[nj][nf];
    jpo->jugada[i].anal.nffpel = analg2.nfipel[nj][nf];
    jpo->jugada[i].anal.ncfin = 0;
    if( jpo->jugada[i].destino.lugar == CORREDOR ){
      n = jpo->jugada[i].destino.pos;
      if( n < ent[nj] ) n += 67;
      jpo->jugada[i].anal.ncfin = n - 17 * nj;
      }
    }
}

/**************************************************************************/

void elige_jugadam( struct Tablero *tbl, struct JugadasPosibles *jpo,
		    struct Jugada *jugada, int *valores )
{
  int i, nj, v, valor = -32000;
//  int valores[4];

  completa_info_anl( tbl,jpo );
  for( i=0; i<jpo->njugadas; i++ ){
    valores[i] = v = analiza_jugada( &(jpo->jugada[i]) );
    if( v > valor ){
      valor = v;
      nj = i;
      }
    }
  *jugada = jpo->jugada[nj];
}

/**************************************************************************/

int analiza_jugada( struct Jugada *jg )
{
  int valor = 0;

  if( jg->anal.sale ) valor += 100;
  if( jg->anal.meta ) valor += 200;
  if( jg->anal.come ) valor += 500;
  if( jg->anal.hpte ) valor += 15;
  if( jg->anal.dseg ) valor -= 20;
  if( jg->anal.aseg ) valor += 20;
  if( jg->anal.dseg && jg->anal.aseg ) valor += 30;
  valor += (jg->anal.nffcom - jg->anal.nficom) * 55;
  valor += (jg->anal.nfipel - jg->anal.nffpel) * 50;
  valor += jg->anal.ncfin / 5;
  valor += random() % 10;  /* un poco de incertidumbre */
  return valor;
}
